from botocore.exceptions import NoCredentialsError, NoRegionError

from dsnap.main import app
import logging

logging.basicConfig(format='[%(levelname)s] %(message)s', level=logging.WARNING)

try:
    app(prog_name="dsnap")
except (NoCredentialsError, NoRegionError) as e:
    logging.error(e.args[0])



import atexit
import json
import logging
import signal

import sys
from typing import cast, TYPE_CHECKING, TypeVar, Iterable

import jmespath
from boto3.resources.collection import ResourceCollection

from dsnap.snapshot import LocalSnapshot
from dsnap.utils import get_name_tag, fatal, cleanup_snap, take_snapshot

from typer import style, colors, secho

if TYPE_CHECKING:
    from mypy_boto3_ec2 import service_resource as r
    from mypy_boto3_ec2 import type_defs as t

INSTANCE_FILTER: 't.FilterTypeDef' = {"Name": 'instance-state-name', "Values": ['running', 'stopped']}


def snaps_from_input(sess, id):
    ec2: 'r.EC2ServiceResource' = sess.resource('ec2')
    if not id:
        for snap in ec2.snapshots.filter(OwnerIds=['self']).all():
            yield snap
    elif id.startswith("i-"):
        i: 'r.Instance' = ec2.Instance(id)
        for d in i.block_device_mappings:
            vol = ec2.Volume(d['Ebs']['VolumeId'])
            for snap in vol.snapshots.all():
                yield snap
    else:
        raise UserWarning(f"Unexpected argument format: {id}, use an instance id or omit the argument to list all snapshots")


def snap_from_input(sess, id) -> 'r.Snapshot':
    """download_from_id is meant to be called from the cli commands and will exit in the case of an error"""
    ec2: 'r.EC2ServiceResource' = sess.resource('ec2')

    if not id:
        inst: 'r.Instance' = resource_prompt(ec2.instances.filter(Filters=[INSTANCE_FILTER]), '[PrivateDnsName, VpcId]')
        vol = resource_prompt(inst.volumes.all(), 'Attachments[*].Device')
        try:
            snap = resource_prompt(cast('r.Volume', vol).snapshots.all(), '[StartTime, OwnerId, Description]')
        except UserWarning:
            snap = ask_to_create_snapshot(vol)
    elif id.startswith('snap-'):
        snap = ec2.Snapshot(id)
    elif id.startswith('i-'):
        vol = resource_prompt(ec2.Instance(id).volumes.all(), 'Attachments[*].Device')
        try:
            snap = resource_prompt(cast('r.Volume', vol).snapshots.all(), '[StartTime, OwnerId, Description]')
        except UserWarning:
            snap = ask_to_create_snapshot(vol)
    else:
        raise UserWarning('unknown argument type, first argument should be an Instance Id or Snapshot Id')

    if not snap:
        raise UserWarning("no snapshot selected")

    return snap


def vol_from_id(sess, i: str) -> 'r.Volume':
    """download_from_id is meant to be called from the cli commands and will exit in the case of an error"""
    ec2: 'r.EC2ServiceResource' = sess.resource('ec2')
    if not i:
        inst: 'r.Instance' = resource_prompt(ec2.instances.filter(Filters=[INSTANCE_FILTER]), '[PrivateDnsName, VpcId]')
        vol: 'r.Volume' = resource_prompt(inst.volumes.all(), 'Attachments[*].Device')
    elif i.startswith('vol-'):
        vol = ec2.Volume(i)
    elif i.startswith('i-'):
        vol = resource_prompt(ec2.Instance(i).volumes.all(), 'Attachments[*].Device')
    else:
        raise UserWarning("unknown argument type, first argument should be an Instance Id or Snapshot Id")

    # vol will be None in cases of invalid argument or no snapshot was selected
    if not vol:
        fatal('Exiting...')

    return vol


def download_snap_id(sess, force, output, snap_id):
    """download_from_id is meant to be called from the cli commands and will exit in the case of an error"""
    secho(f"Selected snapshot with id {style(snap_id, bold=True)}")
    path = (output and output.absolute().as_posix()) or f"{snap_id}.img"
    LocalSnapshot(path, snap_id, boto3_session=sess).fetch(force=force)


T = TypeVar('T')


def item_prompt(resources: Iterable[T], jmespath_msg: str = None) -> T:
    """Prompt's the user for an item to select from the items passed. Item is expected to support the Item protocol."""
    resources = cast(ResourceCollection, resources)

    items = list(resources.all())
    if not resources or len(items) == 0:
        raise UserWarning(f'no items found when calling {resources._py_operation_name}')
    elif len(items) == 1:
        # No need to make a selection if there's only one option
        return list(items)[0]

    msg = ''
    for i, item in enumerate(items):
        if jmespath_msg:
            # dump and load json to convert datetime and similar to something readable
            data = json.loads(json.dumps(item.meta.data, default=str))
            msg = ', '.join(jmespath.search(jmespath_msg, data=data or ''))

        name = get_name_tag(item.tags)
        secho("{}".format(i), bold=True, nl=False, fg=colors.GREEN)

        if name:
            secho(") {} Name: {} ({})".format(style(item.id, underline=True, bold=True), style(name, bold=True), msg))
        else:
            secho(") {} ({})".format(style(item.id, bold=True), msg))

    answer = int(input(style(f'Select {list(items)[0].meta.resource_model.name}:', underline=True) + ' '))

    try:
        return list(items)[answer]
    except IndexError:
        secho(f"Invalid selection, valid inputs are 0 through {len(items) - 1}", file=sys.stderr, fg=colors.RED)
        return item_prompt(resources)


def resource_prompt(resource: 'Iterable[T]', jmespath_msg='') -> T:
    resource = cast('ResourceCollection', resource)
    return item_prompt(resource, jmespath_msg=jmespath_msg)


def ask_to_run(msg, func):
    resp = input(f'{msg} [y/N]: ')
    if resp.lower() == 'y':
        return func()
    return None


def ask_to_create_snapshot(vol: 'r.Volume') -> 'r.Snapshot':
    """Asks the user if we should to create a temporary snapshot.

    If the answer is Y we start snapshot creation, wait for it to finish and register a function to
    delete the snapshot on exit.
    """
    return ask_to_run(style("No snapshots found, create one?", fg=colors.RED), lambda: create_tmp_snap(vol))


def bold(s: str, **kwargs) -> str:
    return style(s, bold=True, **kwargs)


def create_tmp_snap(vol: 'r.Volume') -> 'r.Snapshot':
    """Creates a temporary snapshot that will get deleted when the process exits."""
    secho(f'Creating snapshot from {bold(vol.id)}')
    snap = take_snapshot(vol)

    atexit.register(cleanup_snap(snap))
    signal.signal(signal.SIGTERM, lambda sigs, type: sys.exit())
    print("Waiting for snapshot to complete.")
    snap.wait_until_completed()
    logging.info("Snapshot creation finished")
    return snap



from pathlib import Path
from typing import TYPE_CHECKING, Optional, List

import boto3
import typer
from typer import Option, Typer, secho, style, colors

from dsnap import utils
from dsnap.prompt import snap_from_input, download_snap_id, snaps_from_input, vol_from_id, bold
from dsnap.utils import fatal, take_snapshot

if TYPE_CHECKING:
    from mypy_boto3_ec2 import service_resource as r

app = Typer(name="dsnap", help="Utility for downloading EBS snapshots using the EBS Direct API's.")

sess: boto3.session.Session = boto3.session.Session()

# This gets set via @app.callback before any command runs.
ec2: 'r.EC2ServiceResource' = None  # type: ignore[assignment]


@app.callback()
def session(
        region: str = Option(default='us-east-1', help="Sets the AWS region.", metavar="REGION"),
        profile: str = Option(default=None, help="Shared credential profile to use.", metavar="PROFILE")
):
    global sess, ec2
    sess = boto3.session.Session(region_name=region, profile_name=profile)
    ec2 = sess.resource('ec2')


@app.command()
def init(
        out_dir: Path = typer.Option(Path('.'), help='Output directory to write Vagrantfile'),
        force: bool = typer.Option(False, help='Overwrite any existing Vagrantfile.')
):
    """
    Write out a Vagrantfile template to explore downloaded snapshots.

    If --out-dir is used the given directory will be used instead.
    If --force is used we will overwrite an already present Vagrantfile

    To use the outputed Vagrantfile set the IMAGE environment to the path of the snapshot you want to mount and run vagrant
    up. For example:

    % dsnap init
    % IMAGE=snap-0543a8681adce0086.img vagrant up
    % vagrant ssh
    """
    output = utils.init_vagrant(out_dir, force)
    if output:
        print(f"Wrote Vagrantfile to {style('./'+str(output), bold=True)}")
    else:
        print(f"Vagrantfile already exists at {output}, use the --force to overwrite.")


@app.command("list")
def list_snapshots(
        instance_id: str = typer.Argument(None, help='Optional instance ID to limit listed snapshots to.'),
        devices: List[str] = typer.Option(['/dev/sda', '/dev/xvda'], help='Optional device name to limit snapshots to.'),
):
    """
    List snapshots in AWS.

    If --instance-id is used then snapshots will be limited to that instances default device attachments.
    If --devices is used alongside --instance-id then listed snapshots are for that instances given devices, by default this
    is /dev/sda and /dev/xvda.
    """
    secho("           Id          |   Owneer ID   | Description   ", underline=True)
    try:
        for snap in snaps_from_input(sess, instance_id):
            secho(f"{style(snap.id, bold=True)}   {snap.owner_id}   {snap.description}")
    except UserWarning as e:
        fatal(*e.args)


@app.command()
def get(
        # We use the filename to determine the snapshot id so we can only use directories for the output option.
        output: Path = typer.Option(
            Path('.'),
            file_okay=False,
            dir_okay=True,
            help='If specified output the snapshot to the given directory, the name however is always the snapshot id.',
        ),
        force: bool = typer.Option(False, help='If specified and the snapshot already exists then overwrite it.'),
        ids: Optional[List[str]] = typer.Argument(default=None, help='The remote snapshot ID to fetch.')
):
    """
    Download a snapshot for a given instance or snapshot ID.

    If no Argument is passed then you'll be prompted to select an instance, volume and snapshot to download. If no snapshot
    exists, you can optionally create a temporary one.

    If an instance ID is passed a snapshot for that instance will be downloaded, if more then one exists you'll be prompted
    to select a one.

    If a snapshot ID is passed that snapshot will be downloaded and you will not be prompted for any additional info.
    """
    try:
        if not ids:
            snap = snap_from_input(sess, ids)
            download_snap_id(sess, force, output, snap.id)
        else:
            for id in ids:
                snap = snap_from_input(sess, id)
                download_snap_id(sess, force, output, snap.id)
    except (UserWarning, FileExistsError) as e:
        fatal(*e.args)


@app.command()
def create(ids: List[str] = typer.Argument(
    None,
    help='One or more ID\'s of a instance or volume to create a snapshot for. To avoid being prompted use an explict volume ID'
         ' rather then an instance ID.'
)):
    """
    Create a snapshot for the given instances default device volume.

    The passed argument should be an instance ID, where a snapshot will be created from the default device volume, either
    /dev/sda or /dev/xvda.
    """
    try:
        if not ids:
            fatal("must pass at least one instance or volume id as an argument")
        for i in ids:
            vol = vol_from_id(sess, i)

            devices = ', '.join([a['Device'] for a in vol.attachments])
            instances = ', '.join([a['InstanceId'] for a in vol.attachments])
            secho(f"Creating snapshot from Instance(s): {bold(instances)}, Volume: {bold(vol.id)}, Device: {bold(devices)}")

            s = take_snapshot(vol)
            secho("Created snapshot {} from volume {}".format(bold(s.id), bold(vol.id)), fg=colors.GREEN)

    except UserWarning as e:
        fatal(*e.args)


@app.command()
def delete(ids: List[str] = typer.Argument(None, help='One or more ID\'s of snapshots to delete')):
    """
    Delete a given snapshot.

    The passed argument should be a snapshot ID to delete.
    """
    if not ids:
        fatal("must pass at least one instance id as an argument")
    for i in ids:
        try:
            s = ec2.Snapshot(i)
            s.delete()
            secho(f"Deleted snapshot {style(s.id, bold=True)}")
        except UserWarning as e:
            fatal(*e.args)



import hashlib
import logging
from base64 import b64encode
from pathlib import Path

from typing import List, Iterable, Dict, Optional

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from mypy_boto3_ec2 import service_resource as r


def get_tag(tags: Iterable[Dict[str, str]], key: str) -> str:
    """Takes a list of tags and a key name, returns the the value for the tag with the given key name."""
    if not tags:
        return ''
    name_tag = filter(lambda t: t['Key'] == key, tags)
    return next(map(lambda t: t['Value'], name_tag), '')


def get_name_tag(tags: List[dict]) -> str:
    """Takes a list of tags and returns the value of the Name tag."""
    return get_tag(tags, 'Name')


def cleanup_snap(snap: 'r.Snapshot'):
    def func():
        print(f'Cleaning up snapshot: {snap.id}')
        snap.delete()

    return func


def sha256_check(data: bytes, digest: str) -> bool:
    """Runs sha256 on data and compares it to digest, returns true if these values match.

    digest is expected to be a base64 encoded result of the binary digest.
    """
    m = hashlib.sha256()
    m.update(data)
    chksum = b64encode(m.digest()).decode()
    result = chksum == digest
    if not result:
        logging.error(f'Expected checksum {digest} but got {chksum}')
    return result


def init_vagrant(out_dir: Path = Path('.'), force=False) -> Optional[Path]:
    """Initializes out_dir directory with a templated Vagrantfile for mounting downloaded images"""
    template = Path(__file__).parent.joinpath(Path('files/Vagrantfile'))
    out = out_dir.joinpath(Path('Vagrantfile').name)
    if out.exists() and not force:
        return None
    else:
        out.write_text(template.read_text())
        return out


def fatal(*msg: str):
    logging.fatal('\n'.join(msg))
    exit(1)


def take_snapshot(vol: 'r.Volume', desc: str = '') -> 'r.Snapshot':
    if not desc:
        devices = ', '.join([a['Device'] for a in vol.attachments])
        # volumes can be attached to more then one instance at a time so include all attachments in the description
        instances = ', '.join([a['InstanceId'] for a in vol.attachments])
        desc = "Instance(s): {}, Volume: {}, Device: {}".format(instances, vol.id, devices)

    snap = vol.create_snapshot(
        Description=desc,
        TagSpecifications=[{
            'ResourceType': 'snapshot',
            'Tags': [{'Key': 'dsnap', 'Value': 'true'}]
        }]
    )
    snap.wait_until_completed()
    return snap



import logging
import os
import sys
from pathlib import Path
from queue import Queue, Empty
from threading import Thread
from typing import TYPE_CHECKING, List, Callable

import botocore.config
from botocore.response import StreamingBody

from dsnap.utils import sha256_check

if TYPE_CHECKING:
    from mypy_boto3_ebs.client import EBSClient
    from mypy_boto3_ebs.type_defs import BlockTypeDef

import boto3.resources

MEGABYTE: int = 1024 * 1024
GIGABYTE: int = 1024 * MEGABYTE

RUN_THREADS = 50


class Block:
    client = boto3.client

    def __init__(self, snap: 'Snapshot', resp: 'BlockTypeDef'):
        self.snapshot = snap
        self.BlockIndex = resp['BlockIndex']
        self.Offset: int = resp['BlockIndex'] * snap.block_size_b
        # When using the list_changed_blocks api the process is mostly the same except that we just care about the
        # seecond block token. The first block token would have already been copied over locally and is what we'll be
        # overwriting.
        self.BlockToken = resp['BlockToken']
        self.BlockData: StreamingBody = None  # type: ignore[assignment]
        self.Checksum: str = ''

    def write(self) -> int:
        """Takes a WriteBlock object to write to disk and yields the number of MiB's for each write."""
        logging.debug(f"Writing block at offset {self.Offset}")
        data = self.BlockData.read()

        if not sha256_check(data, self.Checksum):
            raise UserWarning(f"Got block with incorrect checksum at block offset {self.Offset}")

        with os.fdopen(os.open(self.snapshot.path, os.O_RDWR | os.O_CREAT), 'rb+') as f:
            f.seek(self.Offset)
            bytes_written = f.write(data)
            f.flush()
            return bytes_written

    def fetch(self) -> 'Block':
        logging.debug(f"Getting block index {self.BlockIndex}")
        resp = self.snapshot.ebs.get_snapshot_block(
            SnapshotId=self.snapshot.snapshot_id,
            BlockIndex=self.BlockIndex,
            BlockToken=self.BlockToken,
        )
        self.BlockData = resp['BlockData']
        self.Checksum = resp['Checksum']
        return self


class Snapshot:
    def __init__(
            self,
            snapshot_id: str,
            boto3_session: boto3.session.Session = boto3.session.Session(region_name='us-east-1'),
            botocore_conf: botocore.config.Config = botocore.config.Config()
    ) -> None:
        self.blocks: List[Block] = []
        self.snapshot_id = snapshot_id
        self.path = ''

        self.queue: Queue = Queue()

        # Make sure the number of connections matches the number of threads we run when fetching the EBS snapshot
        ebs_config = botocore.config.Config(max_pool_connections=RUN_THREADS).merge(botocore_conf)
        self.ebs: EBSClient = boto3_session.client('ebs', config=ebs_config)

        self.volume_size_b = 0
        self.total_blocks = 0
        self.blocks_written = 0
        self.block_size_b = 0

    def get_blocks(self) -> List[Block]:
        """Retrieves the list of blocks for self.snapshot_id.

        Various attributes are set when calling this method, best to call this early.
        """
        for block in self._get_blocks():
            self.blocks.append(Block(self, block))
        return self.blocks

    def _get_blocks(self) -> List['BlockTypeDef']:
        resp = self.ebs.list_snapshot_blocks(SnapshotId=self.snapshot_id)

        # BlockIndex is equal to 512 KiB and seek uses bytes.
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ebs.html#EBS.Client.put_snapshot_block
        self.block_size_b = resp['BlockSize']
        self.volume_size_b = resp['VolumeSize'] * GIGABYTE
        logging.info(f"Volume size is {self.volume_size_b}")

        blocks = resp['Blocks']
        while resp.get('NextToken'):
            resp = self.ebs.list_snapshot_blocks(SnapshotId=self.snapshot_id, NextToken=resp.get('NextToken'))
            blocks.extend(resp['Blocks'])

        self.total_blocks = len(blocks)
        logging.info(f"Number of blocks in image: {self.total_blocks}")

        return blocks

    def run(self, func: Callable[[Block], None], threads=RUN_THREADS):
        """Calls func on each block passing it a Block object.

        Run's across number of threads passed in `threads`, this defaults to 50.
        """
        for block in self.blocks:
            logging.debug(f"Putting block index {block.BlockIndex} on the queue")
            self.queue.put(block)

        threads = list()
        for i in range(RUN_THREADS):
            t = Thread(target=lambda: self._run(func))
            threads.append(t)
            t.start()

        self.queue.join()
        for t in threads:
            t.join()

    def _run(self, f: Callable[[Block], None]) -> None:
        while True:
            try:
                block: Block = self.queue.get(block=False)
                f(block)
                self.blocks_written += 1
                print(f"Saved block {self.blocks_written} of {self.total_blocks}", end='\r', file=sys.stderr)
                self.queue.task_done()
            except Exception as e:
                if isinstance(e, Empty):
                    return
                else:
                    logging.exception(f"[ERROR] {e.args}")
                raise e


class LocalSnapshot(Snapshot):
    def __init__(
            self,
            dir: str,
            snapshot_id: str,
            boto3_session: boto3.session.Session = boto3.session.Session(region_name='us-east-1'),
            botocore_conf: botocore.config.Config = botocore.config.Config()
    ) -> None:
        super().__init__(snapshot_id, boto3_session, botocore_conf)

        assert dir
        self.path = str(Path(dir).joinpath(f"{snapshot_id}.img"))

    def fetch(self, force: bool = False) -> None:
        """Downloads self.snapshot_id to the self.path.

        If force is true output_file will be overwritten.
        """
        if Path(self.path).exists() and not force:
            raise FileExistsError(f"The output file '{self.path}' already exists.")
        self.path = os.path.abspath(self.path)
        print(f"Output Path: {self.path}")

        self.get_blocks()
        self.truncate()

        def download(b: Block):
            b.fetch().write()
        self.run(download)

    def truncate(self) -> None:
        """Truncates self.output_file to size self.volume_size_b."""
        with open(self.path, 'wb') as f:
            print(f"Truncating file to {self.volume_size_b/GIGABYTE} GB", file=sys.stderr)
            f.truncate(self.volume_size_b)
            f.flush()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dsnap']

package_data = \
{'': ['*'], 'dsnap': ['files/*']}

install_requires = \
['boto3>=1.16.0,<2.0.0', 'urllib3>=1.26.4,<2.0.0']

extras_require = \
{'cli': ['typer>=0.3.2,<0.4.0']}

entry_points = \
{'console_scripts': ['dsnap = dsnap.main:app']}

setup_kwargs = {
    'name': 'dsnap',
    'version': '1.0.0',
    'description': "Utility for downloading EBS snapshots using the EBS Direct API's",
    'long_description': '![Python package](https://github.com/RhinoSecurityLabs/dsnap/workflows/Python%20package/badge.svg?branch=main)\n\n# dsnap\n\nUtility for downloading EBS snapshots using the EBS Direct API\'s.\n\n* Blog Post: [Downloading and Exploring AWS EBS Snapshots](https://rhinosecuritylabs.com/aws/exploring-aws-ebs-snapshots/).\n* This is used as a library by the [Pacu](https://github.com/RhinoSecurityLabs/pacu) module [ebs__download_snapshots](https://github.com/RhinoSecurityLabs/pacu/wiki/Module-Details#ebs__download_snapshots).\n  * Check out the [Pacu integration](#pacu-integration) section to get an idea of the tradeoff\'s between the using this vs the Pacu module.\n\n#### Recording\n\n![Alt Text](./docs/demo.gif)\n\n## Install\n\n### PyPi\n\n```shell\n% pip install -U pip\n% pip install \'dsnap[cli]\'\n```\n\n## Command Reference\n\n```shell\n% dsnap --help\nUsage: dsnap [OPTIONS] COMMAND [ARGS]...\n\n  A utility for managing snapshots via the EBS Direct API.\n\nOptions:\n  --region REGION                 Sets the AWS region.  [default: us-east-1]\n  --profile PROFILE               Shared credential profile to use.\n  --install-completion [bash|zsh|fish|powershell|pwsh]\n                                  Install completion for the specified shell.\n  --show-completion [bash|zsh|fish|powershell|pwsh]\n                                  Show completion for the specified shell, to\n                                  copy it or customize the installation.\n\n  --help                          Show this message and exit.\n\nCommands:\n  create  Create a snapshot for the given instances default device volume.\n  delete  Delete a given snapshot.\n  get     Download a snapshot for a given instance or snapshot ID.\n  init    Write out a Vagrantfile template to explore downloaded snapshots.\n  list    List snapshots in AWS.\n```\n\n## IAM Permissions\n\nNot all these permissions are strictly necessary, for example if you provide an existing snapshot ID with the `get `\ncommand you\'ll only need `ebs:ListSnapshotBlocks` and `ebs:GetSnapshotBlock`. The rest of the permissions are either\nused to find volumes and snapshots based on instance ID\'s, create temporary snapshots if none exist or for use with the\n`create` and `delete` commands.\n\n\n```json\n{\n    "Version": "2012-10-17",\n    "Statement": [\n        {\n            "Effect": "Allow",\n            "Action": [\n                "ebs:ListSnapshotBlocks",\n                "ebs:GetSnapshotBlock",\n                "ec2:DescribeSnapshots",\n                "ec2:DescribeInstances",\n                "ec2:DescribeVolumes",\n                "ec2:DeleteSnapshot",\n                "ec2:CreateSnapshot",\n                "ec2:CreateTags"\n            ],\n            "Resource": "*"\n        }\n    ]\n}\n```\n\n## Examples\n\n### Listing Snapshots\n```shell\n% dsnap list\n           Id          |   Owner ID   |   State\nsnap-0dbb0347f47e38b96   922105094392   completed\n```\n\n### Downloading a Snapshot\n```shell\n% dsnap get snap-0dbb0347f47e38b96\nOutput Path: /cwd/snap-0dbb0347f47e38b96.img\n```\n\nIf you don\'t specify a snapshot  you\'ll get a prompt to ask which one you want to download:\n```shell\n% dsnap get\n0) i-01f0841393cd39f06 (ip-172-31-27-0.ec2.internal, vpc-04a91864355539a41, subnet-0e56cd55282fa9158)\nSelect Instance: 0\n0) vol-0a1aab48b0bc3039d (/dev/sdb)\n1) vol-0c616d718ab00e70c (/dev/xvda)\nSelect Volume: 0\nNo snapshots found, create one? [y/N]: y\nCreating snapshot for Instance(s): i-01f0841393cd39f06 /dev/sdb, Volume: vol-0a1aab48b0bc3039d\nWaiting for snapshot to complete.\nOutput Path: /cwd/snap-0dbb0347f47e38b96.img\nCleaning up snapshot: snap-0543a8681adce0086\n```\n\n### Mounting in Vagrant\nNote: Vagrant does not offer any guarantee\'s of seperation between the guest and the host. It shouldn\'t considered a security boundry like most VM\'s.\n\nThis requires virtualbox to be installed. dsnap init will write a Vagrantfile to the current directory that can be used to mount a specific downloaded snapshot. Conversion to a VDI disk is handled in the Vagrantfile, it will look for the disk file specified in the IMAGE environment variable, convert it to a VDI using `VBoxManage convertdd`. The resulting VDI is destroyed when the Vagrant box is, however the original raw .img file will remain and can be reused as needed.\n\n```shell\n% dsnap init\n% IMAGE=snap-0543a8681adce0086.img vagrant up\n% vagrant ssh\n```\n\n### Mounting With Docker\n\nThis uses libguestfs to work directly with the downloaded img file.\n\n#### Build Docker Container\n```shell\n% git clone https://github.com/RhinoSecurityLabs/dsnap.git\n% cd dsnap\n% make docker/build\n```\n\n#### Run Guestfish Shell\n\n```shell\n% IMAGE=snap-0dbb0347f47e38b96.img make docker/run\n```\n\nThis will take a second to start up. After it drops you into the shell you should be able to run commands like ls, cd, cat. However worth noting they don\'t always behave exactly like they do in a normal shell.\n\nThe output will give you the basics of how to use the guestfish shell. For a full list of command you can run `help --list`.\n\nBelow is an example of starting the shell and printing the contents of /etc/os-release.\n\n```shell\n% IMAGE=snap-0dbb0347f47e38b96.img make docker/run\ndocker run -it -v "/cwd/dsnap/snap-0dbb0347f47e38b96.img:/disks/snap-0dbb0347f47e38b96.img" -w /disks mount --ro -a "snap-0dbb0347f47e38b96.img" -m /dev/sda1:/\n\nWelcome to guestfish, the guest filesystem shell for\nediting virtual machine filesystems and disk images.\n\nType: ‘help’ for help on commands\n      ‘man’ to read the manual\n      ‘quit’ to quit the shell\n\n><fs> cat /etc/os-release\nNAME="Amazon Linux"\nVERSION="2"\nID="amzn"\nID_LIKE="centos rhel fedora"\nVERSION_ID="2"\nPRETTY_NAME="Amazon Linux 2"\nANSI_COLOR="0;33"\nCPE_NAME="cpe:2.3:o:amazon:amazon_linux:2"\nHOME_URL="https://amazonlinux.com/"\n```\n\n## As a Library\n\ndsnap is also meant to be used as a library, however for this purpose it is worth keeping in mind this is an early version and it is still being developed. The interfaces will likely change as new functionality is added.\n\nWe\'ll do our best to make sure we follow SemVer versioning to avoid any breaking changes in minor and patch versions.\n\n## Related tools\n\n### Pacu Integration\n\nThis project is used by [Pacu](https://github.com/RhinoSecurityLabs/pacu) in the\n[ebs__download_snapshots](https://github.com/RhinoSecurityLabs/pacu/wiki/Module-Details#ebs__download_snapshots) module.\nThe primary benefit of using the Pacu module is to reduce unnecessary API call\'s, as a tradeoff it doesn\'t have some\nniceties that are included with dsnap.\n\nFor example the Pacu module reuses snapshots gathered from [ebs__enum_volumes_snapshots](https://github.com/RhinoSecurityLabs/pacu/wiki/Module-Details#ebs__enum_volumes_snapshots), this prevents looking up snapshots more often then needed. At the moment it does not support some dsnap features like creating temporary snapshots or searching for snapshots by instance ID, this however may change in the future.\n\n### Other Interesting Tools\n\nPlease keep in mind that we can\'t vouch for the following tools. However, we felt they were worth mentioning here.\n\n* [ebs-direct-sec-tools](https://github.com/crypsisgroup/ebs-direct-sec-tools)\n  * In particular the [scansecrets](https://github.com/crypsisgroup/ebs-direct-sec-tools#scansecrets) and [diffsecrets](https://github.com/crypsisgroup/ebs-direct-sec-tools#diffsecrets) commands are interesting because instead of mounting the snapshots they scan the raw image for secrets directly.\n\n\n\n## Development\n\nFor CLI development make sure you include the `cli` extra shown below. You\'ll also want to invoke the package by using python\'s `-m` (shown below) for testing local changes, the dnsap binary installed to the environment will only update when you run pip install.\n\n### Setup\n```shell\n% git clone https://github.com/RhinoSecurityLabs/dsnap.git\n% cd dsnap\n% python3 -m venv venv\n% . venv/bin/activate\n% python -m pip install \'.[cli]\'\n```\n\n### Running With Local Changes\n```shell\n% python -m dsnap --help\n```\n\n### Linting and Type Checking\n```shell\n% make lint\n```\n\n### Testing\n```shell\n% make test\n```\n\n',
    'author': 'Ryan Gerstenkorn',
    'author_email': 'ryan.gerstenkorn@rhinosecuritylabs.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
